/*  Emacs style mode select   -*- C -*- */
/* ----------------------------------------------------------------------------- */
/*  $Id:$ */
/*  Copyright (C) 1993-1996 by id Software, Inc. */
/*  This source is available for distribution and/or modification */
/*  only under the terms of the DOOM Source Code License as */
/*  published by id Software. All rights reserved. */
/*  The source is distributed in the hope that it will be useful, */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License */
/*  for more details. */
/*  $Log:$ */
/*  DESCRIPTION: */
/* 	System interface for sound. */
/* ----------------------------------------------------------------------------- */


/* removal of all "internal sound" */
static const char
rcsid[] = "$Id: i_sound.c,v 2.0 1999/06/25 18:20:31 EDT Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include <math.h>

#include <sys/time.h>
#include <sys/types.h>

#ifndef LINUX
#include <sys/filio.h>
#endif

#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>

/*  Linux voxware output. */
#include <linux/soundcard.h>

/*  Timer stuff. Experimental. */
#include <time.h>
#include <signal.h>

#include <sys/ipc.h>
#include <sys/msg.h>


#include "z_zone.h"

#include "i_system.h"
#include "i_sound.h"
#include "m_argv.h"
#include "m_misc.h"
#include "w_wad.h"
#include "d_main.h"
#include "doomdef.h"

extern char soundwadfile[1024];  /*  sound wad file */

int msq_id = -1;

char* musserver_name = "musserver";
#ifdef SNDIPCAUDIOCTL
int snd_id = -1;
#endif /* SNDIPCAUDIOCTL */
/*  Separate sound server process. */
FILE*	sndserver=0;

char*	sndserver_filename = "sndserver";


void I_SetSfxVolume(int volume)
{
	/*  Identical to DOS. */
	/*  Basically, this should propagate */
	/*   the menu/config file setting */
	/*   to the state variable used in */
	/*   the mixing. */
	snd_SfxVolume = volume;
}



/*  Retrieve the raw data lump index */
/*   for a given SFX name. */
int I_GetSfxLumpNum(sfxinfo_t* sfx)
{
	char namebuf[9];
	sprintf(namebuf, "ds%s", sfx->name);
	return W_GetNumForName(namebuf);
}

/*  Starting a sound means adding it */
/*   to the current list of active sounds */
/*   in the internal channels. */
/*  As the SFX info struct contains */
/*   e.g. a pointer to the raw data, */
/*   it is ignored. */
/*  As our sound handling does not handle */
/*   priority, it is ignored. */
/*  Pitching (that is, increased speed of playback) */
/*   is set, but currently not used by mixing. */
int I_StartSound ( int id, int vol, int sep, int pitch, int priority )
{
#ifdef SNDIPCAUDIOCTL
int err = 0;
	if(snd_id != -1) {
		struct msgbuf {
			long msg_type;
			char msg_text[9];
		} msg_buffer;
		
		msg_buffer.msg_type = 6;
		sprintf(msg_buffer.msg_text,"p%2.2x%2.2x%2.2x%2.2x", id, pitch, vol, sep);
		err = msgsnd(snd_id, (struct msgbuf *) &msg_buffer, 9, IPC_NOWAIT);
	} else {
		fprintf(stderr,"Broken sound  server!\n");
	}
#else
	if (sndserver) {
		fprintf(sndserver, "p%2.2x%2.2x%2.2x%2.2x\n", id, pitch, vol, sep);
		fflush(sndserver);
	}
#endif /* SNDIPCAUDIOCTL */
	priority = 0;
	return id;
}

void I_SoundsOff(void)
{
#ifdef SNDIPCAUDIOCTL 
int err=0;
	if(snd_id != -1) {
		struct msgbuf {
			long msg_type;
			char msg_text[9];
		} msg_buffer;
		
		msg_buffer.msg_type = 6;
		sprintf(msg_buffer.msg_text,"kkkkkkkkk");
		err=msgsnd(snd_id, (struct msgbuf *) &msg_buffer, 9, IPC_NOWAIT);
	} else {
		fprintf(stderr,"Broken sound  server!\n");
	}
#else
	if (sndserver) {
		/*  Send a "kill all" command. */
		fprintf(sndserver, "k\n");
		fflush(sndserver);
	}
#endif /* SNDIPCAUDIOCTL */
	/*  Done. */
	return;
}

void I_ShutdownSound(void)
{
#ifdef SNDIPCAUDIOCTL
int err=0;
	I_SoundsOff();
	if(snd_id != -1) {
		struct msgbuf {
			long msg_type;
			char msg_text[9];
		} msg_buffer;
		
		msg_buffer.msg_type = 6;
		sprintf(msg_buffer.msg_text,"qqqqqqqqq");
		err=msgsnd(snd_id, (struct msgbuf *) &msg_buffer, 9, IPC_NOWAIT);
		sleep(1);
		msgctl(snd_id, IPC_RMID, NULL);
	} else {
		fprintf(stderr,"Broken sound  server!\n");
	}
#else
	I_SoundsOff();
	if (sndserver) {
		/*  Send a "quit" command. */
		fprintf(sndserver, "q\n");
		fflush(sndserver);
	}
#endif /* SNDIPCAUDIOCTL */
	sleep(1);
	/*  Done. */
	return;
}


void I_InitSound()
{
	char buffer[2048];
	int i;
	
	i=0;
	sprintf(buffer, "%s",sndserver_filename);
	if ( access(buffer, X_OK) ) {
		if (getenv("DOOMWADDIR")) {
			sprintf(buffer, "%s/%s", getenv("DOOMWADDIR"), sndserver_filename);
		} else {
			sprintf(buffer, "%s/%s", DEVMAPS,sndserver_filename);
		}
	}
	
	/*  start sound process */
	fprintf(stderr, "Sound server [%s]\n", buffer);
	if ( !access(buffer, X_OK) ) {
		strcat(buffer, " -quiet");
		for(i=0;wadfiles[i];i++)  {
			strcat(buffer," -w ");
			strcat(buffer,wadfiles[i]);
		}
		sndserver = popen(buffer, "w");
	} else {
		fprintf(stderr, "Could not start sound server\n");
		exit(1);
	}
#ifdef SNDIPCAUDIOCTL
	snd_id = msgget(53076, IPC_CREAT | 0777);
#endif /* SNDIPCAUDIOCTL */
	sleep(2);
}




/*  MUSIC API. */
FILE *musicp;

void I_InitMusic(void)
{ 
	char *wowee;
	char *bullshit;
	int xy=0;
	int malsiz=0;
	int addsiz=0;
	int i;
	
	if (getenv("DOOMWADDIR")) {
		xy=strlen(getenv("DOOMWADDIR"));
	} else {
		xy=strlen(DEVMAPS);
	}
	malsiz=strlen(musserver_name)+5+xy;
	
	for(i=0;wadfiles[i];i++)  {
		addsiz=strlen(wadfiles[0])+5+addsiz;
	}

	wowee=malloc(malsiz+addsiz);
	bullshit=malloc(malsiz+addsiz);
	sprintf(bullshit,"%s",musserver_name);
	for(xy=0;xy<strlen(bullshit);xy++) {
		if(bullshit[xy]==' ' || bullshit[xy]==0x0a) {
			bullshit[xy]=0x00;
		}
	}

	sprintf(wowee,"%s",musserver_name);
	for(i=0;wadfiles[i];i++) {
		strcat(wowee," -w ");
		strcat(wowee,wadfiles[i]);
	}
	if(!access(bullshit, X_OK)) {
		/* musserver is invoked via a popen() call: */
		musicp=popen(wowee, "w");
		/* I give all permissions I can... */
		msq_id = msgget(53075, IPC_CREAT | 0777);
		fprintf(stderr, "Music server [%s]\n", bullshit);
	} else {
		sprintf(bullshit,"%s/%s",DEVMAPS,musserver_name);
		sprintf(wowee,"%s/%s",DEVMAPS,musserver_name);
		for(xy=0;xy<strlen(bullshit);xy++) {
			if(bullshit[xy]==' ' || bullshit[xy]==0x0a) {
				bullshit[xy]=0x00;
			}
		}
		for(i=0; wadfiles[i]; i++) {
			strcat(wowee," -w ");
			strcat(wowee,wadfiles[i]);
		}
		if(!access(bullshit, X_OK)) {
			/* musserver is invoked via a popen() call: */
			musicp=popen(wowee, "w");
			/* I give all permissions I can... */
			msq_id = msgget(53075, IPC_CREAT | 0777);
			fprintf(stderr, "Music server [%s]\n", bullshit);
		} else {
			fprintf(stderr, "Could not start music server\n");
			exit(1);
		}
	}
	sleep(2);
}

void I_ShutdownMusic(void)
{
	struct msgbuf {
		long msg_type;
		char msg_text[12];
	} msg_buffer;

	if(msq_id != -1) 
		msg_buffer.msg_type = 6;
		sprintf(msg_buffer.msg_text, "qqqqqqqqqqq");
		msgsnd(msq_id, (struct msgbuf *) &msg_buffer, 12, IPC_NOWAIT);
		sleep(4);
		msgctl(msq_id, IPC_RMID, NULL);
	sleep(4);
}

static int	looping=0;
static int	musicdies=-1;

void I_PlaySong(char *music_ressource_string, int looping)
{
	/* printf("music_ressource_string=%s msq_id=%i\n",music_ressource_string,msq_id); */
	musicdies = gametic + TICRATE*30;
	if(msq_id != -1) {
		struct msgbuf {
			long msg_type;
			char msg_text[12];
		} msg_buffer;

		msg_buffer.msg_type = 6;
		msg_buffer.msg_text[0] = 'd';
		sprintf(msg_buffer.msg_text,"d_%s",music_ressource_string);
		msgsnd(msq_id, (struct msgbuf *) &msg_buffer, 12, IPC_NOWAIT);
	} else {
		fprintf(stderr,"Broken music server!\n");
		looping = 0;
		musicdies = 0;
	}
}

void I_SetMusicVolume(int volume)
{
	/*  Internal state variable. */
	snd_MusicVolume = volume;
	/*  Now set volume on output device. */
	if(msq_id != -1) {
		struct msgbuf {
			long msg_type;
			char msg_text[12];
		} msg_buffer;
		msg_buffer.msg_type = 6;
		memset(msg_buffer.msg_text, 0, 12);
		msg_buffer.msg_text[0] = 'v';
		msg_buffer.msg_text[1] = volume;
		msgsnd(msq_id, (struct msgbuf *) &msg_buffer, 12, IPC_NOWAIT);
	}
}

void I_PauseSong (int handle)
{
	/*  UNUSED. */
	handle = 0;
}

void I_ResumeSong (int handle)
{
	/*  UNUSED. */
	handle = 0;
}

void I_StopSong(int handle)
{
	/*  UNUSED. */
	handle = 0;
	
}

/*  Is the song playing? */
int I_QrySongPlaying(int handle)
{
	/*  UNUSED. */
	handle = 0;
	return looping || musicdies > gametic;
}

