/*************************************************************************
 *  musserver.c
 *
 * New code (C) 2006 Andrew J. Kroll <a@oo.ms>
 *
 * Original Copyright (C) 1995 Michael Heasley (mheasley@hmc.edu)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <errno.h>
#include <string.h>
#include "musserver.h"

extern int use_synth;
extern int seqfd;
extern int mixfd;
extern int cachemus(FILE *wadfile, int version);
extern int readmuscache(int lumpcache);

unsigned long lumpsize[35];	/* size in bytes of data lumps */
unsigned long lumppos[35];	/* position of lumps in wad file */
unsigned char *musicdata;
unsigned char *music_cache[35];
unsigned int music_cachesize[35];
int musok=-1;
int doomver=255;
int verbose = 0;
int qid;
int changevol = 1;

int readdir(FILE *wadfile, int *version, int quiet);
int readmus(FILE *wadfile, int lumpnum);
int playmus(char *musdata, unsigned int mussize, int play_once);
void midi_setup(int pref_dev, int dev_num);
void read_genmidi(FILE *wadfile);
void fmload(void);
void cleanup_midi(void);

int silencer=0;
void show_help(void)
{
	fprintf(stderr,"Usage: musserver [options]\n\n");
	fprintf(stderr,"  -d directory     Look in 'directory' for DOOM wad file\n");
	fprintf(stderr,"  -f               Use FM synth device for music playback\n");
	fprintf(stderr,"  -h               Print this message and exit\n");
	fprintf(stderr,"  -l               List detected music devices and exit\n");
	fprintf(stderr,"  -m               Use general midi device for music playback\n");
	fprintf(stderr,"  -t number        Timeout after 'number' seconds when getting IPC message\n");
	fprintf(stderr,"                   queue id\n");
	fprintf(stderr,"  -u number        Use device of type 'number' where 'number' is the type\n");
	fprintf(stderr,"                   reported by 'musserver -l'.  Requires -f or -m option.\n");
	fprintf(stderr,"  -V               Ignore volume change messages from Doom\n");
	fprintf(stderr,"  -v               Verbose\n\n");
	fprintf(stderr,"  -w wadfile       Wad file to use.\n");
}


#define	MAXWADFILES	20
char *wadfiles[MAXWADFILES];
FILE *infile[MAXWADFILES];

void cleanup(int status)
{

	int i;
	struct msqid_ds *dummy;

	cleanup_midi();
	dummy = malloc(sizeof(struct msqid_ds));
	msgctl(qid, IPC_RMID, dummy);
	free(dummy);
	for(i=0;infile[i];i++) {
		fclose(infile[i]);
	}
	fprintf(stderr,"musserver: Exiting.\n");
	exit(status);
}

void Mcleanup(int status)
{

	struct msqid_ds *dummy;

	cleanup_midi();
	dummy = malloc(sizeof(struct msqid_ds));
	msgctl(qid, IPC_RMID, dummy);
	free(dummy);
	fprintf(stderr,"musserver: Exiting.\n");
	exit(status);
}

/*  */
/*  D_AddFile */
/*  */
void D_AddFile (char *file)
{
	int     numwadfiles;
	char    *newfile;
	
	for (numwadfiles = 0 ; wadfiles[numwadfiles] ; numwadfiles++);
	newfile = malloc (strlen(file)+1);
	strcpy (newfile, file);
	wadfiles[numwadfiles] = newfile;
}

void banner(void)
{
	fprintf(stderr,"musserver: MUSic SERVER version %s loaded!\n",MUSSERVER_VERSION);
	fprintf(stderr,"musserver: This version compatable only with 'ajkdoom'\n");
}

int main(int argc, char **argv)
{
	int done = 0;
	int lump = -1;
	int intro = 0;
	int introa = 0;
	int result;
	char *wadfilename;
	char *waddir;
	char *shootwad;
	unsigned int musicsize;
	int x;
	int outtro = 31;
	int opt_dev = 0;
	int list_dev = 0;
	int num_dev = -1;
	int playarg;
	unsigned int timeout = 1500;
	int numwadfiles;
	unsigned int totmem=0;
	
	banner();
	for(numwadfiles=0; numwadfiles<34; numwadfiles++) music_cache[numwadfiles]=malloc(1);
	
	shootwad= malloc(strlen("*****") + 1);
	sprintf(shootwad,"*****");
	waddir = getenv("DOOMWADDIR");
	if (waddir == NULL) {
		waddir=(char *) &WADDIR;
	}
	while ((x = getopt(argc, argv, "d:fhlmt:u:Vvw:")) != -1)
		switch (x) {
			case 'd':
				waddir = optarg;
				break;

			case 'f':
				if (!opt_dev)
					opt_dev = FM_SYNTH;
				else {
					fprintf(stderr,"musserver: You may not specify both the -m and -f options.\n");
					exit(1);
				}
				break;

			case 'h':
				show_help();
				exit(0);
				break;

			case 'l':
				list_dev = 1;
				break;

			case 'm':
				if (!opt_dev)
					opt_dev = EXT_MIDI;
				else {
					fprintf(stderr,"musserver: You may not specify both the -m and -f options.\n");
					exit(1);
				}
				break;

			case 't':
				timeout = 5 * atoi(optarg);
				break;

			case 'u':
				num_dev = atoi(optarg);
				break;

			case 'V':
				changevol = 0;
				break;

			case 'v':
				verbose++;
				break;

			case 'w':
				/* 
				sould really suck in the whole list of wads, 
				then read them in replcaing data as needed 
				since the calling program ALREADY has checked
				that the wad does exist, we will simply accept 
				it, and not do any checks.
				*/
				D_AddFile(optarg);
				/* leave this just to enable the hack for now */
				free(shootwad);
				shootwad = malloc(strlen(optarg) + 1);
				sprintf(shootwad,"%s",optarg);
				break;

			case '?': case ':':
				show_help();
				exit(1);
				break;
		}

	if (list_dev)
		opt_dev = opt_dev + 100;

	if (((opt_dev == 0) || (opt_dev == 100)) && (num_dev != -1)) {
		fprintf(stderr,"musserver: The -u option requires either the -m or the -f option.\n");
		exit(1);
	}

	midi_setup(opt_dev, num_dev);

	/* what should *really* be done is a smarter lookup! */

	/* try the -w option first */
	if(strcmp("*****",shootwad)) {
		/* 
		Okay, some files were offered. this loads the first one, 
		then patches up the rest, and never access them again.
		All data is now cached!
		*/
		
		for (numwadfiles=0; wadfiles[numwadfiles] ; numwadfiles++) {
			fprintf(stderr,"musserver: Checking WADfile '%s'\n", wadfiles[numwadfiles]);

			wadfilename = malloc(strlen(wadfiles[numwadfiles]) + 1);
			sprintf(wadfilename, "%s",wadfiles[numwadfiles]);
			infile[numwadfiles] = fopen(wadfilename, "r");
			if (infile[numwadfiles] == NULL) {
				/* bail out */
				free(shootwad);
				free(wadfilename);
				/* this is fatal */
				fprintf(stderr,"musserver: Game mode indeterminate. Fatal error!\n");
				exit(1);
			}	
			if(doomver<0 || doomver>2) {
				/* ok, which version? readdir automatically picks the proper version... */
				readdir(infile[numwadfiles], &doomver, 2);
			}
		free(wadfilename);
		}
		free(shootwad);
		if(doomver<0 || doomver>2) {
			/* bail out */
			fprintf(stderr,"musserver: Game mode indeterminate. Fatal error!\n");
			exit(1);
		}
		fprintf(stderr,"musserver: Doom version set to %i.\n",doomver);
	} else {
		free(shootwad);
		/* this is fatal */
		fprintf(stderr,"musserver: Game mode indeterminate. Fatal error!\n");
		exit(1);
	}
	if (verbose)
		switch (doomver) {

			case 0:
				fprintf(stderr,"musserver: Playing music for shareware DOOM\n");
				break;

			case 1:
				fprintf(stderr,"musserver: Playing music for registered DOOM\n");
				break;

			case 2:
				fprintf(stderr,"musserver: Playing music for DOOM II\n");
				break;
		}

	switch(doomver) {
		case 0: 
		case 1:
			intro = 28;
			introa = 27;
			break;
		case 2:
			intro = 33;
			introa = 100;
			outtro = 100;
			break;
	}

	/* 
	Cache all lumps!
	Newer stuff loads over the top of the old!
	*/

	for (numwadfiles=0; wadfiles[numwadfiles] ; numwadfiles++) {
		fprintf(stderr,"musserver: Caching music (if any) from '%s'.\n",wadfiles[numwadfiles]);
		cachemus(infile[numwadfiles], doomver);

		if (use_synth) {
			read_genmidi(infile[numwadfiles]);
			if(musok==1) fmload();
		} else {
			musok=1;
		}
	}
	if(musok==-1) {
		fprintf(stderr,"musserver: Couldn't find GENMIDI entry in any wadfile.\n");
		cleanup(1);
	}	

	/* Close fd's, thier no longer used from here on out! */
	for(x=0; infile[x];x++) {
		fclose(infile[x]);
	}

	fprintf(stderr,"musserver: Cache [===================================]\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8");
	for(x=0; x<35;x++) {
		totmem=totmem+music_cachesize[x];
		if(music_cachesize[x]>0) {
			fprintf(stderr,"+");
		} else {
			fprintf(stderr,"-");
		}
	}
	fprintf(stderr,"\nmusserver: Cache size %i Bytes\n",totmem);
	qid = -1;
	x = 0;
	fprintf(stderr,"musserver: Getting message queue id...\n");
	while (qid == -1) {
		x++;
		qid = msgget((key_t)53075, 0);
		if (x > timeout) qid = -2;
		if (qid == -1) {
			if (errno == ENOENT) {
				usleep(200000);
			} else {
				qid = -2;
			}
		}
	}
	if (qid == -2) {
		fprintf(stderr,"musserver: Could not get IPC message queue id.\n");
		Mcleanup(1);
	}

	fprintf(stderr,"musserver: Queue id=%i\n",qid);
	fprintf(stderr,"musserver: Ready.\n");

/* fork to background */
	if (!verbose) {
		if((x = fork()) > 0) {
			exit(0);
		} else { 
			if(x == -1) {
				fprintf(stderr,"musserver: Unable to fork new process.\n");
				perror("fork");
				for(x=0;infile[x];x++) {
					fclose(infile[x]);
				}
				Mcleanup(1);
				exit(1);
			}
		}
	}
	playarg = 2;
	if (verbose) fprintf(stderr,"musserver: Waiting for first message from Doom...\n");
	while (!done) {
		qid = msgget((key_t)53075, 0);
		if ((lump == outtro) || (lump == intro) || (lump == introa))
			playarg = 1;
		else if (lump >= 0)
			playarg = 0;
		else
			lump = intro;
		if ((verbose) && (playarg != 2))
			fprintf(stderr,"Playing music resource number %d\n", lump + 1);
		musicsize = readmuscache(lump);
		result = playmus((char *)musicdata, musicsize, playarg);
		free(musicdata);
		switch (result) {
			case TERMINATED:
				done = 1;
				if (verbose)
					fprintf(stderr,"musserver: Terminated by caller!\n");
				break;
			default:
				if (result >= 500)
					lump = result - 500;
				else {
					done = 1;
					fprintf(stderr,"musserver: Unknown error in music playing.\n");
				}
				break;
		}
		if (playarg == 2)
			playarg = 1;
	}

	Mcleanup(0);
	return(0);
}
