/*************************************************************************
 *  readwad.c
 *
 * New code (C) 2006 Andrew J. Kroll <a@oo.ms>
 *
 * Original Copyright (C) 1995 Michael Heasley (mheasley@hmc.edu)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "musserver.h"

extern unsigned long lumpsize[35];	/* size in bytes of data lumps */
extern unsigned long lumppos[35];	/* positions of lumps in wad file */
extern unsigned char *musicdata;
extern unsigned char *music_cache[35];
extern unsigned int music_cachesize[35];
extern int verbose;
extern struct opl_instr fm_instruments[175];
extern struct sbi_instrument fm_sbi[175];
extern int musok;

void cleanup(int status);


#define WADHEADER 8	/* The wad header is 12 bytes, but we only care about
			   the last 4.  Thus we skip the first 8 */

char *doom2names[] = {
	"D_RUNNIN", "D_STALKS", "D_COUNTD", "D_BETWEE", "D_DOOM", "D_THE_DA",
	"D_SHAWN", "D_DDTBLU", "D_IN_CIT", "D_DEAD", "D_STLKS2", "D_THEDA2",
	"D_DOOM2", "D_DDTBL2", "D_RUNNI2", "D_DEAD2", "D_STLKS3", "D_ROMERO",
	"D_SHAWN2", "D_MESSAG", "D_COUNT2", "D_DDTBL3", "D_AMPIE", "D_THEDA3",
	"D_ADRIAN", "D_MESSG2", "D_ROMER2", "D_TENSE", "D_SHAWN3", "D_OPENIN",
	"D_EVIL", "D_ULTIMA", "D_READ_M", "D_DM2TTL", "D_DM2INT" };

char *doom1names[] = {
	"D_E1M1", "D_E1M2", "D_E1M3", "D_E1M4", "D_E1M5", "D_E1M6", "D_E1M7",
	"D_E1M8", "D_E1M9", "D_E2M1", "D_E2M2", "D_E2M3", "D_E2M4", "D_E2M5",
	"D_E2M6", "D_E2M7", "D_E2M8", "D_E2M9", "D_E3M1", "D_E3M2", "D_E3M3",
	"D_E3M4", "D_E3M5", "D_E3M6", "D_E3M7", "D_E3M8", "D_E3M9", "D_INTROA",
	"D_INTRO", "D_INTER", "D_VICTOR", "D_BUNNY" };

unsigned long genmidipos;

int readdir(FILE *wadfile, int *version, int quiet)
{

	int dv;
	int inmus = 0;		/* flag */
	long count = 0;		/* counting variable */
	long dirpos;		/* position of directory in wad file */
	int lumpindex = 0;	/* index (0 to 34) of current lump */
	char lumpname[8];	/* name of current lump */
	int musiclump;		/* tells whether current lump contains music */
	int wadstate=1;		/* state machine */

	/* skip first 8 bytes to find and read the wad directory offset */
	fseek(wadfile, WADHEADER, SEEK_SET);
	fread((char*)&dirpos, 4, 1, wadfile);

	/* seek to the name of the first wad directory entry */
	fseek(wadfile, dirpos + 8, SEEK_SET);
	dv=*version;

	/* loop to find positions of all the music entries */
	while (!inmus) {
		memset(lumpname,0,8);
		fread(&lumpname, 8, 1, wadfile);
		if (!strncmp(lumpname, "D_", 2)) {
			musiclump = 1;
			if(*version < 0 || *version > 2) {	
			
				if (!strncmp(lumpname, doom1names[0], strlen(doom1names[0]))) dv=0;
				if (!strncmp(lumpname, doom1names[9], strlen(doom1names[9]))) dv=1;
				if (!strncmp(lumpname, doom2names[1], strlen(doom2names[1]))) dv=2;

			} else {
				switch (*version) {
					case 0: case 1:
						for (lumpindex = 0; lumpindex <= 31; lumpindex++) {
							if (!strncmp(lumpname, doom1names[lumpindex], strlen(doom1names[lumpindex]))) {
								wadstate=0;
								break;
							}
							if (lumpindex == 32) {
								if(!quiet) printf("musserver: Unrecognized music entry (%s).\n", lumpname);
								musiclump = 0;
								wadstate=1; /* return error */
							}
						}
						break;
	
					case 2:
						for (; lumpindex <= 34; lumpindex++) {
							if (!strncmp(lumpname, doom2names[lumpindex], strlen(doom2names[lumpindex]))) {
								wadstate=0;
								break;
							}
							if (lumpindex >= 35) {
								if(!quiet) printf("musserver: Unrecognized music entry (%s).\n", lumpname);
								musiclump = 0;
								wadstate=1; /* return error */
							}
						}
					break;
				}
	
				if (musiclump) {
					fseek(wadfile, -16, SEEK_CUR);
					fread((char*)&lumppos[lumpindex], 4, 1, wadfile);
					fread((char*)&lumpsize[lumpindex], 4, 1, wadfile);
					fseek(wadfile, 16, SEEK_CUR);
					lumpindex++;
				}
			}
		} else if (!strncmp(lumpname, "GENMIDI", 7)) {
			fseek(wadfile, -16, SEEK_CUR);
			fread((char*)&genmidipos, 4, 1, wadfile);
			fseek(wadfile, 20, SEEK_CUR);
		} else if (!strcmp(lumpname, "F_END")) {
			break;
		}
		count++;
		if (feof(wadfile)) inmus = 1;
	}
	*version=dv;
	return(wadstate);
}

int readmus(FILE *wadfile, int lumpnum)
{

	struct mus_header {			/* header of music lump */
		char		id[4];
		unsigned short	music_size;
		unsigned short	header_size;
		unsigned short	channels;
		unsigned short	sec_channels;
		unsigned short	instrnum;
		unsigned short	dummy;
	} header;

	fseek(wadfile, lumppos[lumpnum], SEEK_SET);
	fread(&header, 1, sizeof(header), wadfile);

	fseek(wadfile, lumppos[lumpnum], SEEK_SET);

	/* skip instrument data for now */
	fseek(wadfile, header.header_size, SEEK_CUR);

	musicdata = malloc(header.music_size);
	if (musicdata == NULL) {
		printf("musserver: Could not allocate %d bytes for music data.\n", header.music_size);
		cleanup(1);
	}
	fread(musicdata, 1, header.music_size, wadfile);
	return header.music_size;
}

void read_genmidi(FILE *wadfile)
{

	char header[9];

	fseek(wadfile, genmidipos, SEEK_SET);
	fread(&header, 1, 8, wadfile);
	if (strncmp(header, "#OPL_II#", 8)) {
		return;
	}
	fread(&fm_instruments, sizeof(struct opl_instr), 175, wadfile);
	musok=1; /* indicate that atleast one wad had general midi... */
}

/* 
Overlay new data upon old, therefore making the wanted music the default.
*/
int cachemus(FILE *wadfile, int version)
{
	int i;
	int datasize;
	if(verbose < 2) printf("musserver: M-cache [===================================]\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8\x8");
	/* Clear entries */
	for(i=0; i<35; i++) lumpsize[i]=0;
	/* Read in the wad dir... */
	readdir(wadfile,&version,1);
	/* Cache in the new data */
	for(i=0; i<35; i++) {
		if(verbose > 2) printf("musserver: Lump %i, size %li\n",i,lumpsize[i]);  
		if(lumpsize[i]!=0) {
			if(verbose < 2) printf("+");
			/* cache it! */
			free(music_cache[i]);
			datasize=readmus(wadfile,i);
			music_cache[i]=malloc(datasize);
			music_cachesize[i]=datasize;
			memcpy(music_cache[i], musicdata, music_cachesize[i]);
		} else {
			if(verbose < 2) printf("-");
		}
	}
	if(verbose < 2) printf("\n");
	return(0);
}

int readmuscache(int lumpcache)
{
	if(verbose > 2) fprintf(stderr,"musserver: Caller asked for lump #%i.\n",lumpcache); 
	musicdata = malloc(music_cachesize[lumpcache]);
	memcpy(musicdata, music_cache[lumpcache], music_cachesize[lumpcache]);
	return(music_cachesize[lumpcache]);
}
