/*  Emacs style mode select   -*- C++ -*- */
/* ----------------------------------------------------------------------------- */
/*  */
/*  $Id: linux.c,v 1.3 1997/01/26 07:45:01 b1 Exp $ */
/*  */
/*  Copyright (C) 1993-1996 by id Software, Inc. */
/*  */
/*  This source is available for distribution and/or modification */
/*  only under the terms of the DOOM Source Code License as */
/*  published by id Software. All rights reserved. */
/*  */
/*  The source is distributed in the hope that it will be useful, */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License */
/*  for more details. */
/*  */
/*  */
/*  Revision 1.4 1999/04/05 */
/*  More warnings cleaned, more formats supported --AJK */
/*  */
/*  $Log: linux.c,v $ */
/*  Revision 1.3  1997/01/26 07:45:01  b1 */
/*  2nd formatting run, fixed a few warnings as well. */
/*  */
/*  Revision 1.2  1997/01/21 19:00:01  b1 */
/*  First formatting run: */
/*   using Emacs cc-mode.el indentation for C++ now. */
/*  */
/*  Revision 1.1  1997/01/19 17:22:45  b1 */
/*  Initial check in DOOM sources as of Jan. 10th, 1997 */
/*  */
/*  */
/*  DESCRIPTION: */
/* 	UNIX, soundserver for Linux tested only on i386 platform */
/*  */
/* ----------------------------------------------------------------------------- */

static const char rcsid[] = "$Id: linux.c,v 1.3 1997/01/26 07:45:01 b1 Exp $";


#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <errno.h>
#include <linux/soundcard.h>

#include "soundsrv.h"

int	audio_fd;

void
myioctl ( int fd, int command, int* arg )
{
	int rc;
	extern int errno;
	
	rc = ioctl(fd, command, arg);
	if (rc < 0)
	{
		fprintf(stderr, "sndserver: ioctl(dsp,%i,arg) failed\n", command);
		fprintf(stderr, "sndserver: errno=%d\n", errno);
		exit(-1);
	}
}

int I_InitSound ( int samplerate, int samplesize )
{
	
	int i,j;
	
	audio_fd = open("/dev/dsp", O_WRONLY);
	if (audio_fd<0) {
		fprintf(stderr, "sndserver: Could not open /dev/dsp\n");
		fprintf(stderr, "sndserver: Audio is /dev/silence :-)\n");
		i=0;
	} else {
	
		i = 11 | (2<<16);
		myioctl(audio_fd, SNDCTL_DSP_SETFRAGMENT, &i);
	
		myioctl(audio_fd, SNDCTL_DSP_RESET, 0);
		i=11025;
		myioctl(audio_fd, SNDCTL_DSP_SPEED, &i);
		i=1;
		myioctl(audio_fd, SNDCTL_DSP_STEREO, &i);
	
		myioctl(audio_fd, SNDCTL_DSP_GETFMTS, &j);

/*

PROBABLY NEVER SUPPORT
AFMT_MU_LAW	MuLAW
AFMT_A_LAW	ALAW
AFMT_IMA_ADPCM	ADPCM
AFMT_MPEG	MPEG (2) audio


NOT TESTED!
AFMT_S16_BE	Big endian signed 16
AFMT_U16_BE	Big endian U16

SUPPORTED NOW!
AFMT_U8		unsigned 8
AFMT_S16_LE	Little endian signed 16
AFMT_S8		signed 8
AFMT_U16_LE	Little endian U16

*/
		i=j;
		if (i&=AFMT_S16_LE) {
			fprintf(stderr, "sndserver: Play signed 16 bit audio LE\n");
			myioctl(audio_fd, SNDCTL_DSP_SETFMT, &i);
			i=16;
		} else {
			i=j;
			if(i&=AFMT_U16_LE) {
				fprintf(stderr, "sndserver: Play unsigned 16 bit audio LE\n");
				myioctl(audio_fd, SNDCTL_DSP_SETFMT, &i);
				i=-16;
			} else {
				i=j;
				if(i&=AFMT_S8) {
					fprintf(stderr, "sndserver: Play signed 8 bit audio\n");
					myioctl(audio_fd, SNDCTL_DSP_SETFMT, &i);
					i=8;
				} else {
					i=j;
					if(i&=AFMT_U8) {
						fprintf(stderr, "sndserver: Play unsigned 8 bit audio\n");
						myioctl(audio_fd, SNDCTL_DSP_SETFMT, &i);
						i=-8;
					} else {
						i=j;
						if (i&=AFMT_S16_BE) {
							fprintf(stderr, "sndserver: Play signed 16 bit audio BE\n");
							myioctl(audio_fd, SNDCTL_DSP_SETFMT, &i);
							i=17;
						} else {
							i=j;
							if (i&=AFMT_S16_BE) {
								fprintf(stderr, "sndserver: Play signed 16 bit audio BE\n");
								myioctl(audio_fd, SNDCTL_DSP_SETFMT, &i);
								i=-17;
							} else {
								fprintf(stderr, "sndserver: Audio is /dev/silence :-)\n");
								i=0;
							}
						}
					}
				}
			}
		}
	}
	return(i);
}

void I_SubmitOutputBuffer ( void* samples, int samplecount, int resolution )
{
	if(resolution==16 || resolution==-16 || resolution==17 || resolution==-17) {
		write(audio_fd, samples, samplecount*4);
	}
	if(resolution==8 || resolution==-8) {
		write(audio_fd, samples, samplecount*2);
	}
	if(resolution==0) usleep(100);
}

void I_ShutdownSound(void)
{
	if(audio_fd>0) close(audio_fd);
}

