/*  Emacs style mode select   -*- C++ -*- */
/* ----------------------------------------------------------------------------- */
/*  */
/*  $Id: soundsrv.c,v 1.3 1997/01/29 22:40:44 b1 Exp $ */
/*  */
/*  Copyright (C) 1993-1996 by id Software, Inc. */
/*  */
/*  This source is available for distribution and/or modification */
/*  only under the terms of the DOOM Source Code License as */
/*  published by id Software. All rights reserved. */
/*  */
/*  The source is distributed in the hope that it will be useful, */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License */
/*  for more details. */
/*  */
/*  */
/*  $Log: soundsrv.c,v $ */
/*  Revision 1.3  1997/01/29 22:40:44  b1 */
/*  Reformatting, S (sound) module files. */
/*  */
/*  Revision 1.2  1997/01/21 19:00:07  b1 */
/*  First formatting run: */
/*   using Emacs cc-mode.el indentation for C++ now. */
/*  */
/*  Revision 1.1  1997/01/19 17:22:50  b1 */
/*  Initial check in DOOM sources as of Jan. 10th, 1997 */
/*  */
/*  */
/*  DESCRIPTION: */
/* 	UNIX soundserver, run as a separate process, */
/* 	 started by DOOM program. */
/* 	Originally conceived fopr SGI Irix, */
/* 	 mostly used with Linux voxware. */
/*
	Variable slot count now that dynamiclly changes upon need.
	This means no more cut-off sounds, providing the addon wad 
	does _NOT_ exceed 256 sounds at once.
*/

/*  */
/* ----------------------------------------------------------------------------- */


static const char rcsid[] = "$Id: soundsrv.c,v 1.3 1997/01/29 22:40:44 b1 Exp $";

#define VERSION "2.0.1 2006-01-01"

#include <math.h>
#include <sys/types.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <stdlib.h>
#include <malloc.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <string.h>

#include "sounds.h"
#include "soundsrv.h"
#include "wadread.h"
#ifdef SNDIPCAUDIOCTL
#include <signal.h>
#include <sys/ipc.h>
#include <errno.h>
#ifndef __USE_GNU
#define __USE_GNU
#endif
#include <sys/msg.h>
#endif /* SNDIPCAUDIOCTL */


/* I have only one wad file that can obliterate the max channels... */
#define MAXCHN 255


int bits=0;
int lowestchan=MAXCHN;
int proof=MAXCHN+3;

/* IPC stuff */

#ifdef SNDIPCAUDIOCTL
int qid;
#endif /* SNDIPCAUDIOCTL */

/*  */
/*  Department of Redundancy Department. */
/*  */
typedef struct wadinfo_struct
{
	/*  should be IWAD */
	char	identification[4];
	int		numlumps;
	int		infotableofs;
	
} wadinfo_t;


typedef struct filelump_struct
{
	int		filepos;
	int		size;
	char	name[9];
	
} filelump_t;


/*  an internal time keeper */
static int	mytime = 0;

/*  number of sound effects */
int 		numsounds;

/*  longest sound effect */
int 		longsound;

/*  lengths of all sound effects */
int 		lengths[NUMSFX];

/*  mixing buffer */
signed short	mixbuffer[MIXBUFFERSIZE];

/*  file descriptor of sfx device */
int		sfxdevice;

/*  the channel data pointers */
unsigned char*	channels[2+MAXCHN];

/*  the channel step amount */
unsigned int	channelstep[2+MAXCHN];

/*  0.16 bit remainder of last step */
unsigned int	channelstepremainder[2+MAXCHN];

/*  the channel data end pointers */
unsigned char*	channelsend[2+MAXCHN];

/*  time that the channel started playing */
int		channelstart[2+MAXCHN];

/*  the channel handles */
int 		channelhandles[2+MAXCHN];

/*  the channel left volume lookup */
int*		channelleftvol_lookup[2+MAXCHN];

/*  the channel right volume lookup */
int*		channelrightvol_lookup[2+MAXCHN];

/*  sfx id of the playing sound effect */
int		channelids[2+MAXCHN];

int		snd_verbose=1;

int		steptable[256];

int		vol_lookup[128*256];

int		keepchan[2+MAXCHN];

static void derror(char* msg)
{
	fprintf(stderr, "sndserver: error: %s\n", msg);
	exit(-1);
}

/* the original mixer sucked, here is a better one! --AJK */

void checkend(int x, int *dl, int *dr)
{

	unsigned int		sample;

	/* I _SHOULD_ range check the channel #, but we do not! 
	 Should I plan on making this an inline macro? we'll see.
	*/

	if (channels[x]) {
		sample = *channels[x];
		*dl += channelleftvol_lookup[x][sample];
		*dr += channelrightvol_lookup[x][sample];
		channelstepremainder[x] += channelstep[x];
		channels[x] += channelstepremainder[x] >> 16;
		channelstepremainder[x] &= 65536-1;

		if ((channels[x] >= channelsend[x]) && ( keepchan[x]==1 )) {
			channels[x] = (unsigned char *) S_sfx[channelids[x]].data;
			channelstepremainder[x]=0;
			/* loop it fucker! */
		}
		
		if (channels[x] >= channelsend[x]) channels[x] = 0;
	}
}

/* future master mixer */
int lmix()
{
	return(0);
}

int rmix(void)
{
	return(0); 
}

int mix16(void)
{
	int			dl;
	int			dr;
	
	signed short*		leftout;
	signed short*		rightout;
	signed short*		leftend;
	
	int			step;
	int			i;

	leftout = mixbuffer;
	rightout = mixbuffer+1;
	step = 2;
	
	leftend = mixbuffer + SAMPLECOUNT*step;
	
	/*  mix into the mixing buffer */
	while (leftout != leftend) {
		
		dl = 0;
		dr = 0;
		
		for(i=MAXCHN+2; i>=lowestchan; i--) checkend(i, &dl, &dr);
		/* trim as we go */
		if(lowestchan<MAXCHN && channels[lowestchan] == 0) lowestchan++;

		if (dl > 0x7fff) *leftout = 0x7fff;
		else if (dl < -0x8000) *leftout = -0x8000;
		else *leftout = dl;
		
		if (dr > 0x7fff) *rightout = 0x7fff;
		else if (dr < -0x8000) *rightout = -0x8000;
		else *rightout = dr;
		
		leftout += step;
		rightout += step;
		
	}
	return 1;
}

int mix_16(void)
{
	int			dl;
	int			dr;
	
	signed short*		leftout;
	signed short*		rightout;
	signed short*		leftend;
	
	int				step;
	int				i;
	
	leftout = mixbuffer;
	rightout = mixbuffer+1;
	step = 2;
	
	leftend = mixbuffer + SAMPLECOUNT*step;
	
	/*  mix into the mixing buffer */
	while (leftout != leftend) {
		
		dl = 0;
		dr = 0;
		
		for(i=MAXCHN+2; i>=lowestchan; i--) checkend(i, &dl, &dr);
		/* trim as we go */
		if(lowestchan<MAXCHN && channels[lowestchan] == 0) lowestchan++;
		
		dl=dl+0x8000;
		if (dl > 0xffff) *leftout = 0xffff;
		else if (dl < 0) *leftout = 0;
		else *leftout = dl;
		
		dr=dr+0x8000;
		if (dr > 0xffff) *rightout = 0xffff;
		else if (dr < 0) *rightout = 0;
		else *rightout = dr;
		
		leftout += step;
		rightout += step;
		
	}
	return 1;
}

int mix8(void)
{
	int			dl;
	int			dr;
	
	signed char*		leftout;
	signed char*		rightout;
	signed char*		leftend;
	
	int			step;
	int			i;
	
	leftout =(signed char*) mixbuffer;
	rightout =(signed char*) mixbuffer+1;
	step = 2;
	
	leftend =(signed char*) mixbuffer + SAMPLECOUNT*step;
	
	/*  mix into the mixing buffer */
	while (leftout != leftend) {
		
		dl = 0;
		dr = 0;
		

		for(i=MAXCHN+2; i>=lowestchan; i--) checkend(i, &dl, &dr);
		/* trim as we go */
		if(lowestchan<MAXCHN && channels[lowestchan] == 0) lowestchan++;

		dl=dl/256;		
		/*  Has been char instead of short. */
		if (dl > 127) *leftout = 127;
		else if (dl < -128) *leftout = -128;
		else *leftout = dl;
		
		dr=dr/256;
		if (dr > 127) *rightout = 127;
		else if (dr < -128) *rightout = -128;
		else *rightout = dr;
		
		leftout += step;
		rightout += step;
		
	}
	return 1;
}

int mix_8(void)
{
	int			dl;
	int			dr;
	
	unsigned char*		leftout;
	unsigned char*		rightout;
	unsigned char*		leftend;
	
	int			step;
	int			i;
	
	leftout =(unsigned char*) mixbuffer;
	rightout =(unsigned char*) mixbuffer+1;
	step = 2;
	
	leftend =(unsigned char*) mixbuffer + SAMPLECOUNT*step;
	
	/*  mix into the mixing buffer */
	while (leftout != leftend) {
		
		dl = 0;
		dr = 0;
		
		
		for(i=MAXCHN+2; i>=lowestchan; i--) checkend(i, &dl, &dr);
		/* trim as we go */
		if(lowestchan<MAXCHN && channels[lowestchan] == 0) lowestchan++;
		/* 16 -> 8 */

		dl=(0x8000+dl)/256;
		/*  Has been char instead of short. */
		if (dl > 255) *leftout = 255;
		else if (dl < 0) *leftout = 0;
		else *leftout = dl;
		
		dr=(0x8000+dr)/256;
		if (dr > 255) *rightout = 255;
		else if (dr < 0) *rightout = 0;
		else *rightout = dr;
		
		leftout += step;
		rightout += step;
		
	}
	return 1;
}

int mix(void)
{

	/* assume the 16 bit mixers are endian independent! 
	   this is probably a really bad idea, but oh well! */
	int i=1;
	if(bits==0) usleep(10000);
	if(bits==16 || bits==17) i=mix16();
	if(bits==-16 || bits==-17) i=mix_16();
	if(bits==8) i=mix8();
	if(bits==-8) i=mix_8();
	return(i);
} 

void grabdata ( int c, char** v )
{
	int	i;
	char*	name;
	char*	soundwadname;
	int	wadz=0;
	/*  Now where are TNT and Plutonia. Yuck. */
	soundwadname=malloc(1);
	soundwadname[0]=0x00;
	name=malloc(strlen("Awww crap!")+1);
	sprintf(name,"Awww crap!");
	for (i=1 ; i<c ; i++) {
		if (!strcmp(v[i], "-quiet")) {
			snd_verbose = 0;
		}
		if (!strcmp(v[i], "-w")) {
			free(soundwadname);
			soundwadname=malloc(strlen(v[i+1])+1);
			sprintf(soundwadname, "%s", v[i+1]);
			fprintf(stderr, "sndserver: adding '%s'\n",soundwadname);
			if (! access(soundwadname, R_OK | F_OK) ) {
				name=soundwadname;
			} else {
				fprintf(stderr, "sndserver: error: Could not find wadfile anywhere\n");
				exit(-1);
			}
			openwad(name);
			wadz++;
		}
	}
	
	numsounds = NUMSFX;
	longsound = 0;
	
	for (i=1 ; i<NUMSFX ; i++) {
		if (!S_sfx[i].link) {
			S_sfx[i].data = getsfx(S_sfx[i].name, &lengths[i]);
			if (longsound < lengths[i]) longsound = lengths[i];
		} else {
			S_sfx[i].data = S_sfx[i].link->data;
			lengths[i] = lengths[(S_sfx[i].link - S_sfx)/sizeof(sfxinfo_t)];
		}
	}
}

static struct timeval		last={0,0};
/* static struct timeval		now; */

static struct timezone		whocares;

void updatesounds(void)
{
	
	mix();
	I_SubmitOutputBuffer(mixbuffer, SAMPLECOUNT, bits);
	
}

int addsfx ( int sfxid, int volume, int step, int seperation )
{
	static unsigned short	handlenums = 0;
	
	int		i;
	int		rc = -1;
	
	int		oldest;
	int		oldestnum = 0;
	int		slot;
	int		rightvol;
	int		leftvol;
	
	
	oldest=mytime;
	/*  play these sound effects
	    only one at a time
	    I _DO_ want to preserve the brain sound tho :-) 
	    I also should add a flag so as if any of these audios are not 
	    _OURS_ (actor or monster) we preserve them... 
	    I suppose it wouldn't be to difficult to add on a tag as to _WHO_ 
	    activated a sound.
	    Ones that matter are:
	    Players only, these need a player flag (1-8):
	    sfx_sawup
	    sfx_sawidl
	    sfx_sawful
	    sfx_sawhit
	    sfx_stnmov
	    Player or Actor:
	    sfx_pistol <- needs a machine gun flag... this can be player# _OR_ actor#
	    
	*/
	if ( sfxid == sfx_sawup || sfxid == sfx_sawidl || sfxid == sfx_sawful || sfxid == sfx_sawhit || sfxid == sfx_stnmov || sfxid == sfx_pistol) {
		for (i=0 ; i<MAXCHN ; i++) {
			if (channels[i] && (channelids[i] == sfxid)) {
				channels[i] = 0;
				keepchan[i]=0;
			}
		}
	}
	if(sfxid == sfx_bossit && channels[MAXCHN+1]) {
		channels[MAXCHN+1]=0;
		keepchan[MAXCHN+1]=0;
		slot=MAXCHN+1;
	} else {

		for (i = 0 ; i < MAXCHN ; i++) {
			if(channels[i]!=0) {
				if ((channels[oldestnum]!=0) && ((channelstart[i] < oldest) && ( keepchan[i] != 1))) {
					oldestnum = i;
					oldest = channelstart[i];
				}
			} else oldestnum=i;
		}
		if (i == MAXCHN) slot = oldestnum;
		else slot = i;
	}

	keepchan[slot]=0;

	channels[slot] = (unsigned char *) S_sfx[sfxid].data;
	channelsend[slot] = channels[slot] + lengths[sfxid];
	
	if (!handlenums)
	handlenums = 100;
	
	channelhandles[slot] = rc = handlenums++;
	channelstep[slot] = step;
	channelstepremainder[slot] = 0;
	channelstart[slot] = mytime;
	
	/*  (range: 1 - 256) */
	seperation += 1;
	
	/*  (x^2 seperation) */
	leftvol = volume - (volume*seperation*seperation)/(256*256);
	
	seperation = seperation - 257;
	
	/*  (x^2 seperation) */
	rightvol = volume - (volume*seperation*seperation)/(256*256);
	
	/*  sanity check */
	if (rightvol < 0 || rightvol > 127) derror("rightvol out of bounds");
	
	if (leftvol < 0 || leftvol > 127) derror("leftvol out of bounds");
	
	/*  get the proper lookup table piece */
	/*   for this volume level */
	channelleftvol_lookup[slot] = &vol_lookup[leftvol*256];
	channelrightvol_lookup[slot] = &vol_lookup[rightvol*256];
	
	channelids[slot] = sfxid;

	/* ok, if this is the brain stem, and it's a LONG audio sound ( > 65535) set the loop flag! */
	if ((sfxid == sfx_bossit) && (lengths[sfxid] > 65535)) keepchan[slot]=1;

	if(slot<lowestchan) lowestchan=slot;
	if(slot<proof) proof=slot;
	return rc;
	
}


void outputushort(int num)
{
	
	static unsigned char	buff[5] = { 0, 0, 0, 0, '\n' };
	static char*		badbuff = "xxxx\n";
	
	/*  outputs a 16-bit # in hex or "xxxx" if -1. */
	if (num < 0)
	{
		write(1, badbuff, 5);
	}
	else
	{
		buff[0] = num>>12;
		buff[0] += buff[0] > 9 ? 'a'-10 : '0';
		buff[1] = (num>>8) & 0xf;
		buff[1] += buff[1] > 9 ? 'a'-10 : '0';
		buff[2] = (num>>4) & 0xf;
		buff[2] += buff[2] > 9 ? 'a'-10 : '0';
		buff[3] = num & 0xf;
		buff[3] += buff[3] > 9 ? 'a'-10 : '0';
		write(1, buff, 5);
	}
}

void initdata(void)
{
	
	int		i;
	int		j;
	
	int*	steptablemid = steptable + 128;
	
	for (i=0 ;
	i<sizeof(channels)/sizeof(unsigned char *) ;
	i++)
	{
		channels[i] = 0;
	}
	
	gettimeofday(&last, &whocares);
	
	for (i=-128 ; i<128 ; i++)
	steptablemid[i] = pow(2.0, (i/64.0))*65536.0;
	
	/*  generates volume lookup tables */
	/*   which also turn the unsigned samples */
	/*   into signed samples */
	
	for (i=0 ; i<128 ; i++) {
		for (j=0 ; j<256 ; j++) {

			vol_lookup[i*256+j] = (i*(j-128))*16;
		}
	}
}

void quit(void)
{
	fprintf(stderr,"sndserver: Exiting.\n");
	I_ShutdownSound();
	exit(0);
}


/* IPC stuff */

#ifdef SNDIPCAUDIOCTL
int qid;
#else /* SNDIPCAUDIOCTL */
fd_set		fdset;
fd_set		scratchset;
#endif /* SNDIPCAUDIOCTL */



int main ( int c, char** v )
{
	int rc;
	int sndnum;
	int step;
	int vol;
	int sep;
	int i;
	int done = 0;
	int handle = 0;
	int waitingtofinish=0;

#ifdef SNDIPCAUDIOCTL
	int x;
	struct msgbuf *recv;
	struct msqid_ds *dummy;
#else
	int nrc;
	unsigned char	commandbuf[20];
	struct timeval	zerowait = { 0, 0 };
#endif /* SNDIPCAUDIOCTL */
	
	fprintf(stderr, "SouND SERVER version %s loaded!\n",VERSION);
	fprintf(stderr, "This version compatable only with 'ajkdoom'\n");
	/*  get sound data */
	grabdata(c, v);
	
	/*  init any data */
	initdata();
	
	bits=I_InitSound(11025, 16);
	
#ifdef SNDIPCAUDIOCTL
	qid = -1;
	x = 0;
	printf("musserver: Getting message queue id...\n");
	while (qid == -1) {
		x++;
		qid = msgget((key_t)53076, 0);
		fprintf(stderr,"sndserver: Queue id=%d.\n",qid);
		if (x > 1500) qid = -2;
		if (qid == -1) {
			if (errno == ENOENT) {
				usleep(200000);
			} else {
				qid = -2;
			}
		}
	}
	if (qid == -2) {
		fprintf(stderr,"sndserver: Could not get IPC message queue id, exiting.\n");
		quit();
		exit(1);
	}

	/* fork to background */
	if((x = fork()) > 0) {
		exit(0);
	} else {
		if(x == -1) {
			fprintf(stderr,"sndserver: Unable to fork new process, exiting.\n");
			quit();
			exit(1);
		}
	}

	qid = msgget((key_t)53076, 0);
	recv = malloc(sizeof(long) + 20);
#else /* SNDIPCAUDIOCTL */
	FD_ZERO(&fdset);
	FD_SET(0, &fdset);
#endif /* SNDIPCAUDIOCTL */

	/*  parse commands and play sounds until done */
	fprintf(stderr, "sndserver: Ready.\n");
	
	while (!done) {
		mytime++;
		if (!waitingtofinish) {
			do {
#ifdef SNDIPCAUDIOCTL
				/* IPC grabs all the data in one fell swoop :-) */
				/* 0 == command p/q/s	*/
				/* 12 sound number	*/
				/* 34 step (speed)	*/
				/* 56 volume		*/
				/* 78 seperation	*/
				rc=msgrcv(qid, recv, 9, 0, IPC_NOWAIT + MSG_NOERROR);
				if (rc > 0) {
					if (snd_verbose) {
						recv->mtext[10]=0;
						fprintf(stderr, "sndserver: %s\n", &recv->mtext[0]);
					}
					switch(recv->mtext[0]) {
						
						/* play sound */
						case 'p':
							recv->mtext[1] -=
							recv->mtext[1]>='a' ? 'a'-10 : '0';
							recv->mtext[2] -=
							recv->mtext[2]>='a' ? 'a'-10 : '0';
							recv->mtext[3] -=
							recv->mtext[3]>='a' ? 'a'-10 : '0';
							recv->mtext[4] -=
							recv->mtext[4]>='a' ? 'a'-10 : '0';
							recv->mtext[5] -=
							recv->mtext[5]>='a' ? 'a'-10 : '0';
							recv->mtext[6] -=
							recv->mtext[6]>='a' ? 'a'-10 : '0';
							recv->mtext[7] -=
							recv->mtext[7]>='a' ? 'a'-10 : '0';
							recv->mtext[8] -=
							recv->mtext[8]>='a' ? 'a'-10 : '0';
							
							/*      p<snd#><step><vol><sep> */
							sndnum = (recv->mtext[1]<<4) + recv->mtext[2];
							step = (recv->mtext[3]<<4) + recv->mtext[4];
							step = steptable[step];
							vol = (recv->mtext[5]<<4) + recv->mtext[6];
							sep = (recv->mtext[7]<<4) + recv->mtext[8];
							
							handle = addsfx(sndnum, vol, step, sep);
							/*  returns the handle */
							/*      outputushort(handle); */
							break;
						
						/* quit */
						case 'q':
							waitingtofinish = 1; rc = 0;
							break;
							
						/* save audio feature... interesting... --AJK */
						case 's': 
							{
								int fd;
								recv->mtext[3] = 0;
								fd = open((char*)&recv->mtext[1], O_CREAT|O_WRONLY, 0644);
								recv->mtext[1] -= recv->mtext[1]>='a' ? 'a'-10 : '0';
								recv->mtext[2] -= recv->mtext[2]>='a' ? 'a'-10 : '0';
								sndnum = (recv->mtext[1]<<4) + recv->mtext[2];
								write(fd, S_sfx[sndnum].data, lengths[sndnum]);
								close(fd);
							}
							break;
						
						/* kill all pending sounds */
						case 'k':
							for (i=0 ; i<MAXCHN ; i++) {
								channels[i] = 0;
								keepchan[i]=0;
							}
							break;
							
						default:
							fprintf(stderr, "sndserver: Did not recognize command\n");
							break;
					}	
				} else if (rc == -1 ) {
					switch (errno) {
						case EACCES:
							fprintf(stderr,"sndserver: Could not receive IPC message: no read access permission!\n");
							dummy = malloc(sizeof(struct msqid_ds));
							msgctl(qid, IPC_RMID, dummy);
							free(dummy); 
							quit();
							break;

						case EFAULT:
							fprintf(stderr,"sndserver: Could not receive IPC message: memory address is inaccessible.\n");
							dummy = malloc(sizeof(struct msqid_ds));
							msgctl(qid, IPC_RMID, dummy);
							free(dummy); 
							quit();
							break;

						case EIDRM:
							fprintf(stderr,"sndserver: EIDRM error\n");
							dummy = malloc(sizeof(struct msqid_ds));
							msgctl(qid, IPC_RMID, dummy);
							free(dummy); 
							quit();
							break;

						case EINVAL:
							fprintf(stderr,"sndserver: Could not receive IPC message: invalid message size or queue id.\n");
							dummy = malloc(sizeof(struct msqid_ds));
							msgctl(qid, IPC_RMID, dummy);
							free(dummy); 
							quit();
							break;


						case EINTR:
							fprintf(stderr,"sndserver: Could not receive IPC message: received an interrupt signal.\n");
							break;

						case ENOMSG:
							break;

					}
				}
#else /* SNDIPCAUDIOCTL */
				scratchset = fdset;
				rc = select(FD_SETSIZE, &scratchset, 0, 0, &zerowait);
				if (rc > 0)
				{
					/* 	fprintf(stderr, "sndserver: select is true\n"); */
					/*  got a command */
					nrc = read(0, commandbuf, 1);
					
					if (!nrc)
					{
						done = 1;
						rc = 0;
					}
					else
					{
						if (snd_verbose)
						fprintf(stderr, "sndserver: cmd: %c\n", commandbuf[0]);
						
						switch (commandbuf[0])
						{
							/* play sound */
							case 'p':
								read(0, commandbuf, 9);
								if (snd_verbose) {
									commandbuf[9]=0;
									fprintf(stderr, "sndserver: %s\n", commandbuf);
								}
							
								commandbuf[0] -=
								commandbuf[0]>='a' ? 'a'-10 : '0';
								commandbuf[1] -=
								commandbuf[1]>='a' ? 'a'-10 : '0';
								commandbuf[2] -=
								commandbuf[2]>='a' ? 'a'-10 : '0';
								commandbuf[3] -=
								commandbuf[3]>='a' ? 'a'-10 : '0';
								commandbuf[4] -=
								commandbuf[4]>='a' ? 'a'-10 : '0';
								commandbuf[5] -=
								commandbuf[5]>='a' ? 'a'-10 : '0';
								commandbuf[6] -=
								commandbuf[6]>='a' ? 'a'-10 : '0';
								commandbuf[7] -=
								commandbuf[7]>='a' ? 'a'-10 : '0';
							
								/* 	p<snd#><step><vol><sep> */
								sndnum = (commandbuf[0]<<4) + commandbuf[1];
								step = (commandbuf[2]<<4) + commandbuf[3];
								step = steptable[step];
								vol = (commandbuf[4]<<4) + commandbuf[5];
								sep = (commandbuf[6]<<4) + commandbuf[7];
							
								handle = addsfx(sndnum, vol, step, sep);
								/*  returns the handle */
								/* 	outputushort(handle); */
								break;
							
							/* quit */
							case 'q':
								read(0, commandbuf, 1);
								waitingtofinish = 1; rc = 0;
								break;
							
							/* save audio feature... interesting... --AJK */
							case 's': 
								{
									int fd;
									read(0, commandbuf, 3);
									commandbuf[2] = 0;
									fd = open((char*)commandbuf, O_CREAT|O_WRONLY, 0644);
									commandbuf[0] -= commandbuf[0]>='a' ? 'a'-10 : '0';
									commandbuf[1] -= commandbuf[1]>='a' ? 'a'-10 : '0';
									sndnum = (commandbuf[0]<<4) + commandbuf[1];
									write(fd, S_sfx[sndnum].data, lengths[sndnum]);
									close(fd);
								}
								break;
							/* kill all pending sounds */
							case 'k':
								read(0, commandbuf, 1);
								for (i=0 ; i<MAXCHN ; i++) {
									channels[i] = 0;
									keepchan[i]=0;
								}
								break;
							default:
								fprintf(stderr, "sndserver: Did not recognize command\n");
								break;
						}
					}
				} else if (rc < 0) {
					quit();
				}
#endif /* SNDIPCAUDIOCTL */

			} while (rc > 0);
		}
		updatesounds();
		
		if (waitingtofinish) {
			/* need to check and kill off the stem! */
			for(i=MAXCHN+2 ; i>=lowestchan ; i--) {
				if ( channels[i] && (keepchan[i]==1)) {
					keepchan[i]=0;
					channels[i]=0;
				}
			}
			for(i=MAXCHN+2 ; i>=lowestchan && !channels[i] ; i--);
			if (i==MAXCHN-1 || bits==0) done=1;

		}
	}
	
#ifdef SNDIPCAUDIOCTL
	dummy = malloc(sizeof(struct msqid_ds));
	msgctl(qid, IPC_RMID, dummy);
	free(dummy); 
#endif /* SNDIPCAUDIOCTL */
	quit();
	return 0;
}
