/*  Emacs style mode select   -*- C++ -*- */
/* ----------------------------------------------------------------------------- */
/*  */
/*  $Id: wadread.c,v 1.3 1997/01/30 19:54:23 b1 Exp $ */
/*  */
/*  Copyright (C) 1993-1996 by id Software, Inc. */
/*  */
/*  This source is available for distribution and/or modification */
/*  only under the terms of the DOOM Source Code License as */
/*  published by id Software. All rights reserved. */
/*  */
/*  The source is distributed in the hope that it will be useful, */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License */
/*  for more details. */
/*  */
/*  */
/*  $Log: wadread.c,v $ */
/*  Revision 1.3  1997/01/30 19:54:23  b1 */
/*  Final reformatting run. All the remains (ST, W, WI, Z). */
/*  */
/*  Revision 1.2  1997/01/21 19:00:10  b1 */
/*  First formatting run: */
/*   using Emacs cc-mode.el indentation for C++ now. */
/*  */
/*  Revision 1.1  1997/01/19 17:22:51  b1 */
/*  Initial check in DOOM sources as of Jan. 10th, 1997 */
/*  */
/*  */
/*  DESCRIPTION: */
/* 	WAD and Lump I/O, the second. */
/* 	This time for soundserver only. */
/* 	Welcome to Department of Redundancy Department. Again :-). */
/*  */
/* ----------------------------------------------------------------------------- */


static const char rcsid[] = "$Id: wadread.c,v 1.3 1997/01/30 19:54:23 b1 Exp $";



#include <malloc.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>

#include "soundsrv.h"
#include "wadread.h"


int*		sfxlengths;

typedef struct wadinfo_struct
{
	char	identification[4];
	int		numlumps;
	int		infotableofs;
	
} wadinfo_t;

typedef struct filelump_struct
{
	int		filepos;
	int		size;
	char	name[8];
	
} filelump_t;

typedef struct lumpinfo_struct
{
	int		handle;
	int		filepos;
	int		size;
	char	name[8];
	
} lumpinfo_t;



lumpinfo_t*	lumpinfo=NULL;
int		numlumps=0;
int		firsttime=0;

void**		lumpcache;


#define strcmpi strcasecmp


/*  */
/*  Something new. */
/*  This version of w_wad.c does handle endianess. */
/*  */
#ifndef __BIG_ENDIAN__

#define LONG(x) (x)
#define SHORT(x) (x)

#else

#define LONG(x) ((long)SwapLONG((unsigned long) (x)))
#define SHORT(x) ((short)SwapSHORT((unsigned short) (x)))

unsigned long SwapLONG(unsigned long x)
{
	return
	(x>>24)
	| ((x>>8) & 0xff00)
	| ((x<<8) & 0xff0000)
	| (x<<24);
}

unsigned short SwapSHORT(unsigned short x)
{
	return
	(x>>8) | (x<<8);
}

#endif



void strupr (char *s)
{
	while (*s)
	*s++ = toupper(*s);
}

int filelength (int handle)
{
	struct stat	fileinfo;
	
	if (fstat (handle,&fileinfo) == -1)
	fprintf (stderr, "sndserver: Error fstating\n");
	
	return fileinfo.st_size;
}



void openwad(char* wadname)
{
	
	int wadfile;
	int tableoffset;
	int tablelength;
	int tablefilelength;
	int i;
	wadinfo_t header;
	filelump_t* filetable;
	int lumper;
	int startlump;

	wadfile = open(wadname, O_RDONLY);

	if (wadfile < 0) { 
		fprintf(stderr,"sndserver: Could not open wadfile '%s'",wadname);
		return;
	}
	read(wadfile, &header, sizeof header);
	/* We don't handle bare lumps... */
	
	if(strncmp((char *) &header.identification+1,"WAD", 3)) {
		fprintf(stderr,"sndserver: Not a WADfile '%s'",wadname);
		return;
	}
	fprintf(stderr,"sndserver: WADfile type: \"%4.4s\"\n",header.identification);

	if(firsttime==0) {
		firsttime=1;
		lumpinfo = (lumpinfo_t *) malloc(sizeof(lumpinfo_t));
	}	
	startlump=numlumps;
	lumper = LONG(header.numlumps);
	numlumps=numlumps+lumper;
	tableoffset = LONG(header.infotableofs);
	tablelength = lumper * sizeof(lumpinfo_t);
	tablefilelength = lumper * sizeof(filelump_t);
	
	lumpinfo = (lumpinfo_t *) realloc(lumpinfo,numlumps*sizeof(lumpinfo_t));
	filetable = (filelump_t *) malloc(tablefilelength);
	
	/*  get the lumpinfo table */
	lseek(wadfile, tableoffset, SEEK_SET);
	read(wadfile, filetable, tablefilelength);
	
	/*  process the table to make the endianness right and shift it down */
	/* a smarter way would be to do the merge right here! */
	for (i=startlump ; i<numlumps ; i++) {
		strncpy(lumpinfo[i].name, filetable[i-startlump].name, 8);
		lumpinfo[i].handle = wadfile;
		lumpinfo[i].filepos = LONG(filetable[i-startlump].filepos);
		lumpinfo[i].size = LONG(filetable[i-startlump].size);
		/* 
		fprintf(stderr, "sndserver: lump [%.8s] exists in handle %i\n", lumpinfo[i].name, lumpinfo[i].handle );
		*/
	}
	free(filetable);
	/*
	fprintf(stderr, "sndserver: lump count now is %i, handle was %i\n",numlumps,wadfile);
	*/
}

void* loadlump (char* lumpname, int* size )
{
	
	int i;
	void* lump;
	
	for (i=numlumps-1 ; i>0 ; i--) {
		if (!strncasecmp(lumpinfo[i].name, lumpname, 8)) break;
	}
	
	if (i <=0) {
		/*
		fprintf(stderr, "sndserver: Could not find lumpname [%s]\n", lumpname);
		*/
		lump = NULL;
	} else {
		/*
		fprintf(stderr, "sndserver: lump [%.8s] exists in handle %i, size %i\n",lumpinfo[i].name, lumpinfo[i].handle, lumpinfo[i].size);
		*/
		lump = (void *) malloc(lumpinfo[i].size);
		if (lump==NULL) {
			fprintf(stderr, "sndserver: lump not loadable!\n");
		}
		lseek(lumpinfo[i].handle, lumpinfo[i].filepos, SEEK_SET);
		read(lumpinfo[i].handle, lump, lumpinfo[i].size);
		*size = lumpinfo[i].size;
	}
	
	return lump;
	
}

void* getsfx (char* sfxname, int* len )
{
	unsigned char* sfx;
	unsigned char* paddedsfx;
	int i;
	int size;
	int paddedsize;
	char name[20];
	
	sprintf(name, "ds%s", sfxname);
	
	sfx = (unsigned char *) loadlump(name, &size);
	if(sfx != NULL) {
		/*  pad the sound effect out to the mixing buffer size */
		paddedsize = ((size-8 + (SAMPLECOUNT-1)) / SAMPLECOUNT) * SAMPLECOUNT;
		paddedsfx = (unsigned char *) realloc(sfx, paddedsize+8);
		for (i=size ; i<paddedsize+8 ; i++) paddedsfx[i] = 128;
	
		*len = paddedsize;
		return (void *) (paddedsfx + 8);
	} else {
		return(void *)(NULL);
	}
	
}
